""" A ListenerPool simplifies the synchronization of multiple Listener
instances and doubles as an interface to execute custom methods
everytime the Listeners are synchronized. """
import os
import re
import sys
import time
from threading import Thread

from .listener import Listener


def listener_cpp_thread(listener: Listener):
    """Calls the listening loop of the received Listener. Expects to be
    receiving messages from a C++ Producer """
    listener.listen_cpp()

def listener_thread(listener: Listener):
    """Calls the listening loop of the received Listener. """
    listener.listen()


class ListenerPool:
    """Manages multiple Listeners in different threads, handling their
    sincronization and allowing new methods to be bound.

    Bound methods will be executed when the threads are synchronized.

    Syncronization conditions can also be customized.

    Methods
    -------
    bind(method)
        Binds a method to be executed after syncronization
    set_sync_condition(method)
        Sets sync condition as specified method
    is_synced()
        Returns True if the sync condition is met and False otherwise
    start()
        Starts the verification of synchronization of Listeners and
        execution of bound methods
    get_syncer_thread()
        Returns the thread that verifies the synchronization of
        Listeners
    get_topic_names(regex : str) -> list : Static
        Returns all active topics' names that correspond to regex
    """

    def __init__(self, topic_regex: str, cpp: bool=False):
        """
        topic_regex : str
            Regex string. All topics that satisfy this regex will be
            listened to.
        cpp : bool, optional
            If True, expects listeners to be receiving messages from
            a C++ source code. Else, expects regular python messages.
        """
        self.listener_threads = []
        self.listeners = {}
        topic_names = self.get_topic_names(topic_regex)
        thread_func = listener_thread
        if cpp:
            thread_func = listener_cpp_thread
        for topic in topic_names:
            self.listeners[topic] = Listener(topic=topic)
            t = Thread(target=thread_func, args=[self.listeners[topic]], daemon=True)
            self.listener_threads.append(t)
        self.bound_methods = []
        self.sync_condition = self._base_sync_condition
        self.syncer_thread = Thread(target=self._sync_and_execute, daemon=True)

    def _base_sync_condition(self, listeners: dict):
        """Fallback synchronization method """
        return True

    def _sync_and_execute(self):
        """Waits for Listeners to be synced according to synchronization
        condition set with set_sync_condition().
        
        Loops until there are no more bound methods.
        """
        total_time = 0.0
        time_decay_rate = 0.9
        start_time = time.time()
        while len(self.bound_methods) > 0:
            if self.is_synced():
                time_interval = time.time() - start_time
                start_time = time.time()
                total_time = time_decay_rate * total_time\
                             + (1.0 - time_decay_rate) * time_interval
                fps = 1.0/total_time
                for method in self.bound_methods:
                    if method(self.listeners, fps=fps):
                        self.bound_methods.remove(method)
        for listener in self.listeners.values():
            listener.stop()
        for thread in self.listener_threads:
            thread.join()

    def bind(self, method):
        """Binds method to be executed after synchronization in the main
        loop. Bound methods must be able to receive the 'fps' argument.
        
        Parameters
        ----------
        method : function
            Method to be bound. Bound methods always receive as
            arguments all listeners. When bound method returns True, it
            is removed from the bound methods list. 
        """
        self.bound_methods.append(method)

    def set_sync_condition(self, sync_method):
        """Sets sync condition as the specified method

        Parameters
        ----------
        sync_method : function
            Function that receives all Listeners as argument and must
            return a boolean value indicating if the Listeners are
            synced.
        """
        self.sync_condition = sync_method
    
    def is_synced(self) -> bool:
        """Returns if the synchronization condition is fulfilled or not,
        given the state of all Listener instances.
        """
        return self.sync_condition(self.listeners)

    def start(self):
        """ Starts the listen() loop in all Listener instances and
        starts the syncer_thread that waits for Listeners to be
        synchronized and executes all bound methods in an ongoing loop
        until there are no more bound methods.

        Raises
        ------
        Exception
            If there are no bound methods when this method is called.
        """
        if len(self.bound_methods) == 0:
            raise Exception("ListenerPool can't be started with no bound methods")
        else:
            for thread in self.listener_threads:
                thread.start()
            self.syncer_thread.start()

    def get_syncer_thread(self) -> Thread:
        """ Returns the thread that runs the loop that waits for
        synchronization and then executes all bound methods.

        Returns
        -------
        syncer_thread : Thread
        """
        return self.syncer_thread

    @staticmethod
    def get_topic_names(regex: str) -> list:
        """Returns all Kafka topics that correspond to received regex
        
        Returns
        -------
        topics : List[str]
            A list of strings of topic names that satify the regex
        """
        topic_names = Listener.get_active_topics()
        return [topic for topic in topic_names if re.search(regex, topic)]



"""A Listener is a configurable instance of a Confluent Kafka's
Consumer's abstraction. Any data type that can be deserialized by pickle
can be listened by with a Listener. """
import os
import pickle
from threading import Lock

from confluent_kafka import Consumer


class Listener:
    """Listener capable of receiving, through a Kafka Consumer, messages
    serialized using pickle.
    
    Received messages are stored locally when the listen() method is
    called, and their information is mantained until it is recovered by
    using the get_info() method.

    Note that any new message keys will be permanently added to a new
    dictionary field of this Listener until the clean() method is
    called, so be thoughtful of excessive memory usage.

    Methods
    -------
    subscribe(topic: str)
        Subscribes to topic(s)
    listen()
        Listen for new messages in a loop
    listen_cpp()
        Listen for new messages, sent from CPPKafka, in a loop.
    stop()
        Stops listen()'s loop
    has_info(key: str)
        Returns if Listener has information stored with specified key
    get_info(key: str)
        Returns and consumes information stored with specified key
    clean()
        Cleans the information keys
    get_active_topics(**configs) : Static
        Returns active topics of a Consumer with a customizable
        configuration
    """

    def __init__(self, topic: str=None, ip: str="localhost", port: int=9092, **configs):
        """
        topic : str or list, optional
            Topic(s) the consumer will be subscribed to. If none is set,
            doesn't initially subscribe to any topic.
        ip : str, optional
            IP of the Kafka server where messages will be sent to
            (default is localhost)
        port : int, optional
            Port of the Kafka server where messages will be sent to
            (default is 9092)
        **config : dict, optional
            Additional configurations for the Kafka Consumer
        """
        self.lock = Lock()
        self.stop_thread = False
        self.message = {}
        producer_config = {'bootstrap.servers': str(ip) + ':' + str(port), 'group.id': 'mygroup',
                           'auto.offset.reset': 'latest'}
        for key in configs.keys():
            producer_config[key] = configs[key]
        self.consumer = Consumer(producer_config)
        if topic:
            self.subscribe(topic)
    
    def subscribe(self, topic: str):
        """Subscribes listener to a topic. 
        
        Parameters
        ----------
        topic : str or list
            Kafka topic(s) from which the message will be listened from
        """
        if isinstance(topic, str):
            self.consumer.subscribe([topic])
        elif isinstance(topic, list):
            self.consumer.subscribe(topic)

    def listen_for_message(self):
        """Returns messages recieved by this Listener's Consumer """
        return self.consumer.poll(1.0)

    def listen(self):
        """Listens for information incoming from topic specified in
        initiation. """
        while not self.stop_thread:
            msg = self.listen_for_message()
            if msg is not None:
                self.deserialize(msg.value())

    def listen_cpp(self):
        """Listen for information coming from a Producer in C++.
        
        Expects messages to be not serialized by pickle, but be
        equivalently separated in Kafka's message keys. """
        while not self.stop_thread:
            msg = self.listen_for_message()
            if msg is not None:
                key = str(msg.key(), "utf-8").strip('\0')
                self.message[key] = msg.value()
    
    def stop(self):
        """ Stops listen()'s loop """
        self.stop_thread = True

    def deserialize(self, message):
        """Deserializes message """
        deserialized = pickle.loads(message)
        for key in deserialized.keys():
            self.message[key] = deserialized[key]
    
    def has_info(self, key: str):
        """Returns if this Listener has a message stored with the
        specified key. In case the key is not found or it's contents in
        the message dictionary is None, returns False. """
        return (key in self.message) and (self.message[key] is not None)

    def get_info(self, key: str):
        """Returns the message with the specified key stored in this
        Listener and sets it to None afterwards. """
        response = None
        self.lock.acquire()
        if self.has_info(key):
            response = self.message[key]
            self.message[key] = None
        self.lock.release()
        return response

    def clean(self):
        """Cleans the message variable of all it's keys and values,
        setting it to an empty dictionary """
        self.lock.acquire()
        self.message = {}
        self.lock.release()

    @staticmethod
    def get_active_topics(**configs):
        """ Returns all active topics from a Kafka Consumer
        
        Parameters
        ----------
        **config : dict, optional
            Additional configurations for the Kafka Consumer
        """
        producer_config = {'bootstrap.servers': 'localhost:9092', 'group.id': 'mygroup',
                           'auto.offset.reset': 'latest'}
        for key in configs.keys():
            producer_config[key] = configs[key]
        consumer = Consumer(producer_config)
        return consumer.list_topics().topics.keys()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kafka_messaging",
    version="1.1.1",
    author="Andre Ferrari Moukarzel",
    author_email="andremoukarzel@gmail.com",
    description="Simple interface for sending messages through Kafka",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/zaitt_computer_vision/kafka_api",
    packages=setuptools.find_packages(),
    install_requires=["confluent_kafka"],
    license="MIT",
    python_requires=">=3.6"
)